>From 8ce864afd98bc669db856d7c75f0177981a46fd6 Mon Sep 17 00:00:00 2001
From: Alvaro Herrera <alvherre@alvh.no-ip.org>
Date: Wed, 27 Nov 2013 17:47:16 -0300
Subject: [PATCH 3/5] Don't try to set InvalidXid as page pruning hint

If a transaction updates/deletes a tuple just before aborting, and a
concurrent transaction tries to prune the page concurrently, the pruner
may see HeapTupleSatisfiesVacuum return HEAPTUPLE_DELETE_IN_PROGRESS,
but a later call to HeapTupleGetUpdateXid() return InvalidXid.  This
would cause an assertion failure in development builds, but would be
otherwise Mostly Harmless.

Fix by checking whether the updater Xid is valid before trying to apply
it as page prune point.

Reported by Andres in 20131124000203.GA4403@alap2.anarazel.de
---
 src/backend/access/heap/pruneheap.c |   21 +++++++++++++++------
 1 file changed, 15 insertions(+), 6 deletions(-)

diff --git a/src/backend/access/heap/pruneheap.c b/src/backend/access/heap/pruneheap.c
index 3ec10a0..fdfa37c 100644
--- a/src/backend/access/heap/pruneheap.c
+++ b/src/backend/access/heap/pruneheap.c
@@ -479,13 +479,22 @@ heap_prune_chain(Relation relation, Buffer buffer, OffsetNumber rootoffnum,
 				break;
 
 			case HEAPTUPLE_DELETE_IN_PROGRESS:
+				{
+					TransactionId	xmax;
+
+					/*
+					 * This tuple may soon become DEAD.  Update the hint field
+					 * so that the page is reconsidered for pruning in future.
+					 * If there was a MultiXactId updater, and it aborted after
+					 * HTSV checked, then we will get an invalid Xid here.
+					 * There is no need for future pruning of the page in that
+					 * case, so skip it.
+					 */
+					xmax = HeapTupleHeaderGetUpdateXid(htup);
+					if (TransactionIdIsValid(xmax))
+						heap_prune_record_prunable(prstate, xmax);
+				}
 
-				/*
-				 * This tuple may soon become DEAD.  Update the hint field so
-				 * that the page is reconsidered for pruning in future.
-				 */
-				heap_prune_record_prunable(prstate,
-										   HeapTupleHeaderGetUpdateXid(htup));
 				break;
 
 			case HEAPTUPLE_LIVE:
-- 
1.7.10.4

